class tcpBaseException (Exception):
    """
    All TCP exceptions inherit from this exception.
    """

class tcpHttpBaseException (tcpBaseException):
    """
    All tcp HTTP Exceptions inherit from this exception.
    """

    def __init__ (self, *args, **kwargs):
        for key, val in kwargs.items():
            setattr(self, key, val)
        super(tcpHttpBaseException, self).__init__(*args)

class HttpClientError (tcpHttpBaseException):
    """
    Called when the server tells us there was a client error (4xx)
    """

class HttpServerError (tcpHttpBaseException):
    """
    Called when the server tells us there was a server error (5xx)
    """

class InvalidCredentials (tcpHttpBaseException):
    """
    Invalid or no credentials were provided.
    """

class NoDocumentation (tcpHttpBaseException):
    """
    No documentation is available for this endpoint.
    """

class DownloadError (tcpHttpBaseException):
    """
    Download has failed somehow.
    """

class UploadError (tcpHttpBaseException):
    """
    Upload has failed somehow.
    """



import platform
import requests
import sys

from slumber.serialize import Serializer
from slumber.exceptions import HttpClientError, HttpServerError

from .clientAPI import clientAPI
from . import exceptions 
from ._version import __version__

class client (object):

    user_agent = 'scw-sdk/%s Python/%s %s' % (__version__, ' '.join(sys.version.split()), platform.platform())

    def __init__ (self, host:str=None, token:str=None, user_agent:str=None, usermail:str=None, passwd:str=None):

        import os

        if host == None:
            if 'TCP_HOST' in os.environ.keys ():
                host = os.environ['TCP_HOST']
            else:
                host = "https://api.thecrossproduct.xyz/v1"

        self.host = host

        if user_agent:
            self.user_agent = user_agent

        self.token = None

        if token:
            self.token = token

        elif usermail and passwd:

            from requests.auth import HTTPBasicAuth
            import json
            from .logs import warning

            resp = clientAPI (self.host, 
                              self.host,
                              session=self._make_requests_session(),
                              auth=HTTPBasicAuth(usermail,passwd),
                           ).auth.login.get()

            self.token = resp['token']

        if not self.token:
            if 'TCP_API_TOKEN' not in os.environ.keys():
                raise exceptions.InvalidCredentials ("No token available: either use BasicAuth or set the env var $TCP_API_TOKEN") 

            self.token = os.environ['TCP_API_TOKEN']

    def _make_requests_session (self):

        session = requests.Session ()

        session.headers.update({'User-Agent': self.user_agent})

        if self.token:

            session.headers.update ({'Authorization': f'Bearer {self.token}'})

        return session

    def query (self, **kwargs):

        api = clientAPI (self.host,
                         self.host,
                         session=self._make_requests_session(),
                         serializer=Serializer(default="json"),
                         **kwargs)

        return api

    def _get_endpoints (self):

        api = clientAPI (self.host,
                         self.host + '/help',
                         session=self._make_requests_session(),
                         serializer=Serializer(default="json"))

        return api._get_resource(**api._store).get()

    def help (self):

        import textwrap

        resp = self._get_endpoints ()

        lines = []
        for endpoint in resp:
            lines.append ([" OR ".join(endpoint['methods']), 'query()'+endpoint["endpoint"].replace('/', '.')])

        max_first = max([len(x[0]) for x in lines])

        print ("Python SDK to query TCP's API.\n"
               "\n"
               "You can connect to your TCP account by either setting the environment variable TCP_API_TOKEN\n"
               "Alternatively, you can connect using the following code:\n"
               "\n"
               "import tcp\n"
               "client = tcp.client (usermail=\"user@mail.co\", passwd=\"passwd\")\n"
               "\n"
               "If you're looking to send a GET HTTP request against our API, like:\n"
               "\n"
               " GET https://{hostname}/{vX}/auth\n"
               "\n"
               "You only need to call the following pythonic code:\n"
               "\n"
               "import tcp\n"
               "client = tcp.client ()\n"
               "client.query().auth.get()\n"
               "\n"
               "The query method returns a slumber.API object.\n"
               "The latter handles all the excruciating details of the requests.\n"
               "\n"
              )

        print ("The following endpoints are available:\n")

        for line in lines:
            print ('{0:<30}\t{1:<}'.format(line[0], line[1]))

        print ("\n\nOther methods includes:\n"
               "\n"
               "help\t\t- this message\n"
               "download\t- from TCP S3 storage to your local storage\n"
               "upload\t\t- from your local storage to TCP S3 storage\n"
               "dashboard\t- interactive overview of your account ressources")

    def upload (self, src_local:str, dest_s3:str, max_part_size:str=None):
        '''
        Multipart upload of a file from local repository to S3 repository.

        Args:
            src_local (str): path to your file on your local computer
            dest_s3 (str): desired path in TCP S3 bucket
            max_part_size (str): optional. Size of each part to be sent. Either an int (number of bytes) or a human formatted string (example: "10Gb") 

        Exceptions:
            tcp.exceptions.tcpUploadException

        Notes:

            Works accordingly to the following sequence:
                1. Defining the target space in the S3 repository.
                2. Partionning the file and sending each part to the target space
                3. Merging files and completing the upload

            Internally it uses the following TCP endpoints:
                - POST generate_presigned_multipart_post 
                - POST complete_multipart_post
        '''
        import os

        #TODO adding multithread and retry process when error
        #1: S3 target space definition
        file_size = str(os.path.getsize(src_local))
        presigned_body={"uri": dest_s3,"size": file_size}

        if max_part_size:
            presigned_body.update({"part_size":max_part_size})
        
        try:
            resp=self.query().data.generate_presigned_multipart_post.post(presigned_body)
        except slumber.exceptions.SlumberHttpBaseException as err:
            raise exceptions.UploadException (str(err), err.__dict__)

        #2: File's parts loading
        uploadId=resp["upload_id"]
        urls=resp["parts"]
        part_size = resp["part_size"]
        completed_parts=[]

        has_failed = False

        with open(src_local, 'rb') as f:
            for part_no,url in enumerate(urls):
                file_data = f.read(int(part_size))
           
                resp=requests.put(url, data=file_data)

                if resp.status_code != 200:
                    has_failed = True
                    break
                                
                completed_parts.append({'ETag': resp.headers['ETag'].replace('"',''), 'PartNumber': part_no+1})

        body = {}

        # A part has failed, we abort.
        if has_failed:
            body['upload_id'] = uploadId
            body['uri'] = dest_s3
            self.query().data.abort_multipart_post.post (body)

            number_of_parts_done = len(completed_parts)
            total_number_of_parts = len(urls)

            raise exceptions.UploadException (f"Part number {number_of_parts_done} failed ({total_number_of_parts} in totals). The multi-part upload was aborted")

        #3: Parts concatenation and end of upload
        body["upload_id"] = uploadId
        body["parts"] = completed_parts
        body["uri"] = dest_s3

        try:
            self.query().data.complete_multipart_post.post(body)
        except slumber.exceptions.SlumberHttpBaseException as err:
            raise exceptions.UploadException (str(err), err.__dict__)

    def download (self, src_s3, dest_local, chunk_size=8192):
        '''
        Download of a file from local repository to S3 repository.

        Args:
            src_s3 (str): path in TCP S3 bucket 
            dest_local (str): desired path in your local computer 
            chunk_size (int): desired chunk size for streaming download

        Exceptions:
            tcp.exceptions.tcpDownloadException

        Notes:

            Works accordingly to the following sequence:
                1. Get a temporary link to our S3 
                2. Download file from that link using a stream 

            Internally it uses the following TCP endpoints:
                - POST generate_presigned_get
        '''

        body = {} 
        body['uri'] = src_s3

        try:
            resp = self.query ().data.generate_presigned_get.post (body)
        except slumber.exceptions.SlumberHttpBaseException as err:
            raise exceptions.DownloadException (str(err), err.__dict__)

        url = resp['url']

        try: 
            with requests.get(url, stream=True) as r:
                r.raise_for_status ()
                with open (dest_local, 'wb') as f:
                    for chunk in r.iter_content(chunk_size=chunk_size):
                        f.write (chunk)
        except requests.exceptions.HTTPError as err:
            raise exceptions.DownloadException (str(err), err.__dict__)

    def dashboard (self, refresh_delay=5):
        '''
        Print an overview of all processes.

        Args:
            refresh_delay (int): refresh every X seconds
        '''

        import prettytable
        import time
        import curses
        import traceback
        import threading, queue

        ignores_fields = {'Process':['user_id', 'endpoint', 'terminated', 'expires'],
                          'Remote':['user_id', 'usr', 'input_path', 'output_path', 'working_path'],
                          'Instance':['user_id', 'expires', 'ip', 'ssh_usr', 'input_path', 'working_path', 'output_path', 'num_cores', 'mem_required', 'ram_required'],
                          'License':['user_id', 'created', 'last_failed', 'active']
                          }
            
        models = [(self.query().app.list("Process"), "Process"), 
                  (self.query().app.list("Process"), "Instance"), 
                  (self.query().app.list("Process"), "Remote"),
                  (self.query().lics, "License")]

        text = { "Process": ["No Process"],
                 "Instance": ["No Instance"],
                 "Remote": ["No Remote"],
                 "License": ["No License"]}

        current = "Process"
        current_line = 0

        text_queue = queue.Queue()
        text_queue.put (text)

        dims = [80, 80]

        stop_updating = threading.Event ()

        def update_tables (text_queue):

            def get_table (endpoint, model):

                entries = endpoint.get()

                out = [f"No {model}"]

                if entries:

                    filtered = [{key: entry[key] for key in entry if key not in ignores_fields[model]} for entry in entries]

                    table = prettytable.PrettyTable()
                    #table.max_table_width = table.min_table_width = dims[1]
                    #table.max_table_heigth = table.min_table_heigth = dims[0]-5

                    table.field_names = filtered[0].keys ()
                    table.add_rows ([x.values() for x in filtered])

                    out = table.get_string ().split('\n')
                
                return out

            while True:

                text_queue.put ( { model: get_table(endpoint, model) for endpoint, model in models} )

                if stop_updating.is_set ():
                   break 

                time.sleep (refresh_delay)

        th = threading.Thread (target=update_tables, args=(text_queue,))
        th.start ()

        try:
            stdscr = curses.initscr()
            curses.noecho ()
            curses.cbreak ()
            curses.halfdelay(10*refresh_delay)
            stdscr.keypad(1)

            should_exit = False

            current_pos = 0

            while not should_exit:
    
                dims = stdscr.getmaxyx()

                if dims[0] < 40 and dims[1] < 40:
                    stdscr.addstr (0, 0, "screen should be at least 40x40...")
                    continue

                sections = []
                for model in [m[1] for m in models]:
                    if model == current:
                        sections.append (f'<{model}>')
                    else:
                        sections.append (model)

                footnote = '|'.join(sections)+'\n'
                footnote += 'Press q to quit. Press up and down arrow to scroll. Press left and right to select which table.'

                if text_queue.qsize() > 0:
                    text = text_queue.get ()
    
                stdscr.clear ()
                stdscr.addstr(0, 0, '\n'.join(text[current][current_line:min(len(text[current]), dims[0]-5)]))
                stdscr.addstr(dims[0]-2, 0, footnote)

                stdscr.refresh ()

                try:
                    ch = stdscr.getch ()
                except:
                    pass                 

                if ch == curses.KEY_LEFT:
                    index_current = list([m[1] for m in models]).index(current)
                    index_current = max(0, index_current-1)
                    current = models[index_current][1]
                    current_line = 0
                elif ch == curses.KEY_RIGHT:
                    index_current = list([m[1] for m in models]).index(current)
                    index_current = min (len(models)-1, index_current+1)
                    current = models[index_current][1]
                    current_line = 0
                elif ch == curses.KEY_UP:
                    current_line = max(0, current_line-1)
                elif ch == curses.KEY_DOWN:
                    current_line = min(max(0,(len(text[current])-dims[0]+5)), current_line+1)
                elif ch == ord('q'):
                    should_exit = True

                curses.flushinp()

            stdscr.clear ()
            stdscr.addstr ("Exiting...")
            stdscr.refresh ()

        finally:

            stop_updating.set ()
            th.join()

            stdscr.keypad(0)
            curses.echo(); curses.nocbreak()
            curses.endwin()
            traceback.print_exc()




import datetime
import unittest
import tcp
import re

class DataTestCase (unittest.TestCase):

    def setUp (self):

        import os

        self._test_account = os.environ["TCP_TEST_ACCOUNT"]
        self._test_passwd = os.environ["TCP_TEST_PASSWD"]
        self._test_dir = '.'

        self._client = tcp.client (usermail=self._test_account, passwd= self._test_passwd)

    def test_get (self):

        resp = self._client.query().data.get()

        groups = [x['name'] for x in self._client.query().auth.get()['group']]

        assert isinstance(resp, dict)
        assert 'files' in resp
        assert isinstance(resp['files'], list)
        for group in groups:
            assert group in resp
            assert isinstance(resp[group], list)

    def test_download_upload_delete (self):

        import os
        import filecmp

        file_src = os.path.join (self._test_dir, 'test.txt')
        file_dest = os.path.join (self._test_dir, 'test-downloaded.txt')

        with open(file_src, 'w') as f:
            f.write ("This is a test. You can safely remove this file.")

        self._client.upload (file_src, 'test.txt')
        self._client.download ('test.txt', file_dest)

        assert os.path.exists (file_dest)

        filecmp.cmp (file_src, file_dest)

        os.remove (file_src)
        os.remove (file_dest)

        self._client.query().data.delete_file.post ({'uri':"test.txt"})

        resp = self._client.query().data.get()
        assert 'test.txt' not in resp['files']

if __name__ == '__main__':
    unittest.main()



import datetime
import unittest
import tcp
import re

class AuthTestCase (unittest.TestCase):

    def setUp (self):

        import os

        self._test_account = os.environ["TCP_TEST_ACCOUNT"]
        self._test_passwd = os.environ["TCP_TEST_PASSWD"]

        self._client = tcp.client (usermail=self._test_account, passwd= self._test_passwd)

        self._re_jwt = "^[A-Za-z0-9-_]*\.[A-Za-z0-9-_]*\.[A-Za-z0-9-_]*$"
        self._re_uuid4 = "^[0-9a-f]{8}\-[0-9a-f]{4}\-4[0-9a-f]{3}\-[89ab][0-9a-f]{3}\-[0-9a-f]{12}$"
        self._re_mail = "^[\w\.-]+@([\w-]+\.)+[\w-]{2,4}$"

    def test_login_get (self):

        client = tcp.client (usermail=self._test_account, passwd= self._test_passwd)

        assert isinstance(client.token, str)
        assert len(client.token)>0
        assert re.fullmatch("^[A-Za-z0-9-_]*\.[A-Za-z0-9-_]*\.[A-Za-z0-9-_]*$", client.token)

    def test_login_post (self):

        body = {"mail":self._test_account, "passwd":self._test_passwd}
        resp = self._client.query().auth.login.post(body)
        token = resp['token']

        assert isinstance(token, str)
        assert len(token)>0
        assert re.fullmatch(self._re_jwt, token)

    def test_validate_auth_token_get (self):

        resp = self._client.query().auth.validate_auth_token.get()

        assert isinstance (resp, dict)

        for key in resp.keys():
            assert key in ["id", "user_id", "expires_at", "limit", "uses", "scope"]

        assert isinstance(resp["id"], str)
        assert re.fullmatch(self._re_uuid4, resp["id"]) 
        assert isinstance(resp["user_id"], str)
        assert re.fullmatch(self._re_uuid4, resp["user_id"])
        assert isinstance(resp["expires_at"], str)
        datetime.datetime.fromisoformat (resp['expires_at'])
        assert isinstance(resp["limit"], int)
        assert resp["limit"] >= 0
        assert isinstance(resp["uses"], int)
        assert resp["uses"] >= 0
        assert isinstance(resp["scope"], str)

    def test_get (self):

        resp = self._client.query().auth.get ()

        assert isinstance (resp, dict)

        for key in resp:
            assert key in ['id',
                           'mail',
                           'created_on',
                           'nbr_activations',
                           'last_activation',
                           'group',
                           'vat',
                           'address',
                           'company',
                           'contact']

        assert isinstance(resp["id"], str)
        assert re.fullmatch(self._re_uuid4, resp["id"]) 
        assert isinstance(resp["mail"], str)
        assert re.fullmatch(self._re_mail, resp["mail"])
        assert isinstance(resp["created_on"], str)
        datetime.datetime.fromisoformat (resp['created_on'])
        assert isinstance(resp["nbr_activations"], int)
        assert resp["nbr_activations"] >= 0
        assert isinstance(resp["last_activation"], str)
        datetime.datetime.fromisoformat (resp['last_activation'])
        assert isinstance(resp["group"], list)
        for group in resp["group"]:
            assert isinstance(group, dict)
            for key in group.keys():
                assert key in ["id", "name", "descr"]
            assert isinstance(group["id"], str)
            assert re.fullmatch(self._re_uuid4, group["id"])
            assert isinstance(group["name"], str)
            assert isinstance(group["descr"], str)
        assert resp["vat"] is None
        assert resp["address"] is None
        assert resp["company"] is None
        assert resp["contact"] is None

    def test_logout_get (self):

        client = tcp.client (usermail=self._test_account, passwd= self._test_passwd)
        client.query().auth.logout.get()

        try:
            client.query().auth.get()
        except tcp.exceptions.InvalidCredentials as err:
            assert err.response.status_code == 401
            pass

if __name__ == '__main__':
    unittest.main()



import time
import slumber
from .logs import warning
from . import exceptions

class clientResource (slumber.Resource):

    MAX_RETRIES = 8
    MAX_DELAY = 32

    def _get_help (self):
        new_uri = self._store["host"] + '/help' + self._store["base_url"].replace(self._store["host"], "")
        self._store.update({"base_url": new_uri} )

        try:
            resp = self.get ()
        except exceptions.HttpClientError as err:
            if err.response.status_code == "404":
                raise exceptions.NoDocumentation(str(err), **err.__dict__) from err
            raise err 

        try:
            return resp.decode('utf-8')
        except UnicodeDecodeError as err:
            raise exceptions.NoDocumentation (f"Unable to decode") from err
        return ''

    def help (self):
        print (self._get_help())

    def _retry_in (self, retry):

        return min(2 ** retry, self.MAX_DELAY)

    def _request (self, *args, **kwargs):

        retry = 0

        while True:
            try:
                return super(clientResource, self)._request(*args, **kwargs)
            except slumber.exceptions.HttpClientError as err:
                if err.response.status_code == 401:
                    raise exceptions.InvalidCredentials (str(err), **err.__dict__) from err
                raise exceptions.HttpClientError (str(err), **err.__dict__) from err
            except slumber.exceptions.HttpServerError as err:
                if err.response.status_code not in (502, 503, 504):
                    raise exceptions.HttpServerError (str(err), **err.__dict__) from err

            retry += 1
            retry_in = self._retry_in(retry)

            if retry >= self.MAX_RETRIES:
                raise exceptions.HttpServerError(f'API endpoint still in maintenance after {retry} attempts.'
                                                      'Stop trying.')

            warning ( 'API endpoint is currently in maintenance. Try again in '
                     f'{retry_in} seconds... (retry {retry} on {self.MAX_RETRIES})'
                     )
            time.sleep (retry_in)

    def _process_response (self, resp):
        if not self._store.get('serialize', True):
            return resp

        return super (clientResource, self)._process_response(resp)

class clientAPI (slumber.API):
    resource_class = clientResource

    def __init__ (self, host=None, base_url=None, auth=None, format=None, append_slash=True, session=None, serializer=None):

        super ().__init__ (base_url, auth, format, append_slash, session, serializer)
        self._store.update ({"host": host})



import datetime
import unittest
import tcp
import re

class LicsTestCase (unittest.TestCase):

    def setUp (self):

        import os

        self._test_account = os.environ["TCP_TEST_ACCOUNT"]
        self._test_passwd = os.environ["TCP_TEST_PASSWD"]

        self._client = tcp.client (usermail=self._test_account, passwd= self._test_passwd)

        self._re_uuid4 = "^[0-9a-f]{8}\-[0-9a-f]{4}\-4[0-9a-f]{3}\-[89ab][0-9a-f]{3}\-[0-9a-f]{12}$"
        self._re_datetime = "^(Mon|Tue|Wed|Thu|Fri|Sat|Sun)[ ]+(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)[ ]+([1-9]|[1-2][0-9]|3[0-1])[ ]+([0-1][0-9]|2[0-4]):([0-5][0-9]|60):([0-5][0-9]|60)[ ]+(1|[0-9]{4})$"

    def test_get (self):

        resp = self._client.query().lics.get()

        assert isinstance (resp, list)

        for lic in resp:
            assert isinstance (lic, dict)

            for key in lic:
                assert key in ['token',
                               'name',
                               'user_id',
                               'scope',
                               'created',
                               'until',
                               'last_used',
                               'last_failed',
                               'active',
                               'uses',
                               'limit',
                               'failed_attempts']

            assert isinstance (lic['token'], str)
            assert re.fullmatch (self._re_uuid4, lic['token'])
            assert isinstance (lic['name'], str)
            assert isinstance (lic['user_id'], str)
            assert re.fullmatch (self._re_uuid4, lic['user_id'])
            assert isinstance (lic['scope'], str)
            assert isinstance (lic['created'], str)
            datetime.datetime.fromisoformat (lic['created'])
            assert isinstance (lic['until'], str)
            datetime.datetime.fromisoformat (lic['until'])
            assert isinstance (lic['last_used'], str)
            datetime.datetime.fromisoformat (lic['last_used'])
            assert isinstance (lic['last_failed'], str)
            datetime.datetime.fromisoformat (lic['last_failed'])
            assert isinstance (lic['active'], bool)
            assert isinstance (lic['uses'], float)
            assert lic['uses'] >= 0
            assert isinstance (lic['limit'], float)
            assert lic['limit'] >= 0
            assert isinstance (lic['failed_attempts'], int)
            assert lic['failed_attempts'] >= 0

if __name__ == '__main__':
    unittest.main()



import datetime
import unittest
import tcp
import re

class AppTestCase (unittest.TestCase):

    def setUp (self):

        import os

        self._test_account = os.environ["TCP_TEST_ACCOUNT"]
        self._test_passwd = os.environ["TCP_TEST_PASSWD"]

        self._client = tcp.client (usermail=self._test_account, passwd= self._test_passwd)

        self._re_uuid4 = "^[0-9a-f]{8}\-[0-9a-f]{4}\-4[0-9a-f]{3}\-[89ab][0-9a-f]{3}\-[0-9a-f]{12}$"
        self._re_ip = "^((25[0-5]|(2[0-4]|1\d|[1-9]|)\d)\.){3}(25[0-5]|(2[0-4]|1\d|[1-9]|)\d)$"
        self._re_usr = "^[a-z_]([a-z0-9_-]{0,31}|[a-z0-9_-]{0,30}\$)$"
        self._re_path = "^([^ !$`&*()+]|(\\[ !$`&*()+]))+$"
        self._re_mem = "^[1-9][0-9]{0,32}(|.[0-9]+)(|b|Kb|Mb|Gb|Tb|Pb)$"
        self._re_remote_cp = "^(scw|aws):[a-zA-Z0-9_-]+:[\.a-zA-Z0-9_-]+$"
        self._re_remote_id = "^(scw|aws):[a-zA-Z0-9_-]+:[\.a-zA-Z0-9_-]+:[\.a-zA-Z0-9_-]+$"

    def test_get (self):
        resp = self._client.query().app.get()

        assert isinstance (resp, dict)

        for key in resp:
            assert isinstance (resp[key], list)
            for el in resp[key]:
                assert isinstance (el, str)

    def test_list_Process_get (self):

        resp = self._client.query().app.list.Process.get() 

        assert isinstance (resp, list)

        for el in resp:

            for key in el:
                assert key in ['id', 
                               'user_id', 
                               'app',
                               'domain',
                               'endpoint',
                               'launched',
                               'terminated',
                               'expires',
                               'state']

            assert isinstance (el, dict) 
            assert isinstance (el['id'], str)
            assert re.fullmatch (self._re_uuid4, el['id'])
            assert isinstance (el['user_id'], str)
            assert re.fullmatch (self._re_uuid4, el['user_id'])
            assert isinstance (el['app'], str)
            assert isinstance (el['domain'], str)
            assert isinstance (el['endpoint'], str)
            assert el['endpoint'] in ['run', 'quotation']
            assert isinstance (el['launched'], str)
            datetime.datetime.fromisoformat (el['launched'])
            assert isinstance (el['terminated'], str)
            datetime.datetime.fromisoformat (el['terminated'])
            assert isinstance (el['expires'], str)
            datetime.datetime.fromisoformat (el['expires'])
            assert isinstance (el['state'], str)

    def test_list_Instance_get (self):

        resp = self._client.query().app.list.Instance.get() 

        assert isinstance (resp, list)

        for el in resp:

            for key in el:
                assert key in ['id', 
                               'ext_id', 
                               'user_id', 
                               'process_id',
                               'launched',
                               'expires',
                               'state',
                               'ip',
                               'ssh_usr',
                               'input_path',
                               'working_path',
                               'output_path',
                               'num_cores',
                               'mem_required',
                               'ram_required']

            assert isinstance (el, dict) 
            assert isinstance (el['id'], str)
            assert re.fullmatch (self._re_uuid4, el['id'])
            assert isinstance (el['ext_id'], str)
            assert (re.fullmatch ('^remote:'+self._re_uuid4[1:], el['ext_id']) or re.fullmatch(self._re_remote_id, el['ext_id']) or not el['ext_id'])
            assert isinstance (el['user_id'], str)
            assert re.fullmatch (self._re_uuid4, el['user_id'])
            assert isinstance (el['process_id'], str)
            assert re.fullmatch (self._re_uuid4, el['process_id'])
            assert isinstance (el['launched'], str)
            datetime.datetime.fromisoformat (el['launched'])
            assert isinstance (el['expires'], str)
            datetime.datetime.fromisoformat (el['expires'])
            assert isinstance (el['state'], str)
            assert isinstance (el['ip'], str) 
            assert re.fullmatch (self._re_ip, el['ip']) or not el['ip']
            assert isinstance (el['ssh_usr'], str)
            assert re.fullmatch (self._re_usr, el['ssh_usr']) or not el['ssh_usr']
            assert isinstance (el['input_path'], str)
            assert re.fullmatch (self._re_path, el['input_path']) or not el['input_path']
            assert isinstance (el['working_path'], str)
            assert re.fullmatch (self._re_path, el['working_path']) or not el['working_path']
            assert isinstance (el['output_path'], str)
            assert re.fullmatch (self._re_path, el['output_path']) or not el['output_path']
            assert isinstance (el['num_cores'], int)
            assert el['num_cores'] > 0
            assert isinstance (el['mem_required'], int)
            assert el['mem_required'] >= 0
            assert isinstance (el['ram_required'], int)
            assert el['ram_required'] >= 0

    def test_list_Remote_get (self):

        resp = self._client.query().app.list.Remote.get() 

        assert isinstance (resp, list)

        for el in resp:

            for key in el:
                assert key in ['id', 
                               'name', 
                               'user_id', 
                               'ip',
                               'usr',
                               'num_cores',
                               'mem',
                               'ram',
                               'input_path',
                               'working_path',
                               'output_path',
                               'instanciated']

            assert isinstance (el, dict) 
            assert isinstance (el['id'], str)
            assert (re.fullmatch ('^remote:'+self._re_uuid4[1:], el['id']) or re.fullmatch(self._re_remote_id, el['id']))
            assert isinstance (el['user_id'], str)
            assert re.fullmatch (self._re_uuid4, el['user_id'])
            assert isinstance (el['ip'], str) 
            assert re.fullmatch (self._re_ip, el['ip'])
            assert isinstance (el['usr'], str)
            assert re.fullmatch (self._re_usr, el['usr'])
            assert isinstance (el['num_cores'], int)
            assert el['num_cores'] > 0
            assert isinstance (el['mem'], int)
            assert el['mem'] >= 0
            assert isinstance (el['ram'], int)
            assert el['ram'] >= 0
            assert isinstance (el['input_path'], str)
            assert re.fullmatch (self._re_path, el['input_path'])
            assert isinstance (el['working_path'], str)
            assert re.fullmatch (self._re_path, el['working_path'])
            assert isinstance (el['output_path'], str)
            assert re.fullmatch (self._re_path, el['output_path'])
            assert isinstance (el['instanciated'], bool)

    def test_list_PostMortem_get (self):

        resp = self._client.query().app.list.PostMortem.get() 

        assert isinstance (resp, list)

        for el in resp:

            for key in el:
                assert key in ['id', 
                               'ext_id', 
                               'user_id', 
                               'process_id',
                               'launched',
                               'terminated',
                               'state',
                               'num_cores',
                               'mem',
                               'ram']

            assert isinstance (el, dict) 
            assert isinstance (el['id'], str)
            assert re.fullmatch (self._re_uuid4, el['id'])
            assert isinstance (el['ext_id'], str)
            assert (re.fullmatch ('^remote:'+self._re_uuid4[1:], el['ext_id']) or re.fullmatch(self._re_remote_id, el['ext_id']))
            assert isinstance (el['user_id'], str)
            assert re.fullmatch (self._re_uuid4, el['user_id'])
            assert isinstance (el['process_id'], str)
            assert re.fullmatch (self._re_uuid4, el['process_id'])
            assert isinstance (el['launched'], str)
            datetime.datetime.fromisoformat (el['launched'])
            assert isinstance (el['terminated'], str)
            datetime.datetime.fromisoformat (el['terminated'])
            assert isinstance (el['state'], str)
            assert isinstance (el['num_cores'], int)
            assert el['num_cores'] > 0
            assert isinstance (el['mem'], int)
            assert el['mem'] >= 0
            assert isinstance (el['ram'], int)
            assert el['ram'] >= 0

    def test_remote_create_get_delete (self):
      
        import uuid

        first = 'test_remote_'+str(uuid.uuid4()).replace ('-','_')
        second = 'test_remote_'+str(uuid.uuid4()).replace ('-','_')

        for ii, name in enumerate([first, second]):

            body = {
                    "name": name,
                    "ip": "127.0.0.1",
                    "usr": "test_user",
                    "mem": "10Gb",
                    "ram": "10Gb",
                    "num_cores" : ii+1,
                    "input_path": "/home/test_user/data/input",
                    "working_path": "/home/test_user/data/working",
                    "output_path": "/home/test_user/data/output"
                }
    
            resp = self._client.query().app.new_remote.post (body)
    
            assert isinstance (resp, dict)
            for key in resp:
                assert key in ['id', 'pub']
    
            assert isinstance (resp['id'], str)
            assert re.fullmatch ("^remote:"+self._re_uuid4[1:], resp['id'])
            assert isinstance (resp['pub'], str)
            assert re.fullmatch ("^ssh-ed25519 [a-zA-Z0-9\/+]+$", resp['pub'])

        resp = self._client.query().app.remote.get ()

        assert isinstance (resp, dict)

        for remote_name in [first, second]:
            assert remote_name in resp
            assert isinstance(resp[remote_name], dict)

            remote = resp[remote_name]
    
            for key in remote:
                assert key in ['id', 'num_cores', 'mem', 'ram', 'instanciated']
    
            assert isinstance (remote['id'], str)
            assert re.fullmatch("^remote:"+self._re_uuid4[1:], remote['id'])
            assert isinstance (remote['num_cores'], int)
            assert remote['num_cores'] > 0
            assert isinstance (remote['mem'], str)
            assert re.fullmatch(self._re_mem, remote['mem'])
            assert isinstance (remote['ram'], str)
            assert re.fullmatch(self._re_mem, remote['ram'])
            assert isinstance (remote['instanciated'], bool)
            assert not remote['instanciated']

        resp = self._client.query().app.remote.post({"cores":2})

        assert first not in resp
        assert second in resp

        assert self._client.query().app.remote(first).delete ()
        assert self._client.query().app.remote(second).delete ()

    def check_remote_cp_get_post (self, cloud_providor, creds=None):

        if creds:
            resp = self._client.query().app.remote(cloud_providor).get ()
        else:
            resp = self._client.query().app.remote(cloud_providor).post (creds)

        assert isinstance(resp, dict)

        # Asserting only the first tenth elements
        max_element = min (len(resp), 10)
        keys = list(resp.keys())[:max_element]
        remotes = {}
        for ii in range(max_element):
            remotes[keys[ii]] = resp[keys[ii]]

        for key in remotes:
            remote = remotes[key]
            assert re.fullmatch (self._re_remote_cp, key)
            assert isinstance (remote, dict)
            for kk in remote:
                assert kk in ['cp', 'price', 'ram', 'cores']
            assert remote['cp'] in ['scw', 'aws']
            assert isinstance(remote['price'], dict)
            for key in remote['price']:
                assert key in ["AUD",
                               "BRL",
                               "BGN",
                               "CAD",
                               "CNY",
                               "HRK",
                               "CYP",
                               "CZK",
                               "DKK",
                               "EEK",
                               "EUR",
                               "HKD",
                               "HUF",
                               "ISK",
                               "IDR",
                               "JPY",
                               "KRW",
                               "LVL",
                               "LTL",
                               "MYR",
                               "MTL",
                               "NZD",
                               "NOK",
                               "PHP",
                               "PLN",
                               "RON",
                               "RUB",
                               "SGD",
                               "SKK",
                               "SIT",
                               "ZAR",
                               "SEK",
                               "CHF",
                               "THB",
                               "TRY",
                               "GBP",
                               "USD"]
                assert isinstance (remote['price'][key], float)
            assert isinstance(remote['ram'], str)
            assert re.fullmatch (self._re_mem, remote['ram'])
            assert isinstance(remote['cores'], int)            
            assert remote['cores'] > 0

    def test_remote_scw_get_post (self):
        return self.check_remote_cp_get_post ('scw')

    def test_remote_aws_get_post (self):
        import os 

        for field in ["AWS_ACCESS_KEY_ID",
                      "AWS_SECRET_ACCESS_KEY",
                      "AWS_S3_BUCKET",
                      "AWS_REGION"]
            assert field in os.environ 

        creds = {
            "aws": {
                "access_key_id": os.environ["AWS_ACCESS_KEY_ID"],
                "secret_access_key": os.environ["AWS_SECRET_ACCESS_KEY"],
                "s3_bucket": os.environ["AWS_S3_BUCKET"],
                }
            } 

        return self.check_remote_cp_get_post ('aws', creds)

    def test_datacenters_get_post (self):

        resp = self._client.query().app.datacenters("scw").get ()

        assert isinstance (resp, list)
        for el in resp:
            assert isinstance (el, str)

        for dc in ["fr-par-1",
                   "fr-par-2",
                   "fr-par-3",
                   "nl-ams-1",
                   "nl-ams-2",
                   "pl-waw-1",
                   "pl-waw-2"]:

            assert dc in resp

    def test_info_get (self):

        resp = self._client.query().app.info.get(Domain="test", App="helloworld")

        assert isinstance (resp, bytes)

    def process (self, body):

        resp = self._client.query().app.run.post(body, Domain="test", App="helloworld")

        assert isinstance (resp, dict)
        assert len(resp) == 1
        assert 'id' in resp
        assert isinstance (resp['id'], str)
        assert re.fullmatch (self._re_uuid4, resp['id'])

        puid = resp['id']

        # Should be a process in the process list. We can test it
        self.test_list_Process_get ()

        def check_status ():

            resp = self._client.query().app.status.get (Process=puid)

            assert isinstance (resp, dict)
            for key in resp:
                assert key in ['user_id',
                               'app',
                               'domain',
                               'endpoint',
                               'id',
                               'launched',
                               'terminated',
                               'expires',
                               'state',
                               'outputs']
    
            assert isinstance (resp['user_id'], str)
            assert re.fullmatch (self._re_uuid4, resp['user_id'])
            assert resp['app'] == 'helloworld'
            assert resp['domain'] == 'test'
            assert resp['endpoint'] == 'run'
            assert isinstance (resp['id'], str)
            assert re.fullmatch (self._re_uuid4, resp['id'])
            assert isinstance (resp['launched'], str)
            datetime.datetime.fromisoformat (resp['launched'])
            assert isinstance (resp['terminated'], str)
            datetime.datetime.fromisoformat (resp['terminated'])
            assert isinstance (resp['expires'], str)
            datetime.datetime.fromisoformat (resp['expires'])
            assert isinstance (resp['state'], str)
            assert resp['state'] in ['say_hello', 'upload', 'pending', 'waiting', 'dead', 'terminated']
            assert isinstance (resp['outputs'], dict)

            if resp['outputs']:
                for key in resp['outputs']:
                    assert isinstance (resp['outputs'][key], str)
                    assert resp['outputs'][key].startswith('https://s3')
       
            return resp['state']

        import time

        status = "None"
        num_tries, num_tries_max = 0, 60
        sleep_duration = 10

        while status != "dead":

            if num_tries == num_tries_max:
                break

            num_tries += 1

            time.sleep (sleep_duration)

            status = check_status ()

        resp = self._client.query().app.status.get (Process=puid)

        assert resp['state'] == 'dead'

        assert 'test/msg-test.txt' in resp['outputs']

        resp = self._client.query().app.cost.get (Process=puid)

        assert isinstance (resp, dict)
        assert len(resp) == 1
        assert 'cost' in resp
        assert isinstance (resp['cost'], float)
        assert resp['cost'] < 1.0

    def test_scaleway_process (self):

        pool = ["scw:fr-par-1:PLAY2-PICO"]        
        body = {
                "inputs": {},
                "output-prefix": "test",
                "pool": pool 
            }

        self.process (body) 

#    def test_aws_process (self):
#
#        import os 
#
#        for field in ["AWS_ACCESS_KEY_ID",
#                      "AWS_SECRET_ACCESS_KEY",
#                      "AWS_S3_BUCKET",
#                      "AWS_REGION"]
#            assert field in os.environ 
#
#        creds = {
#            "aws": {
#                "access_key_id": os.environ["AWS_ACCESS_KEY_ID"],
#                "secret_access_key": os.environ["AWS_SECRET_ACCESS_KEY"],
#                "s3_bucket": os.environ["AWS_S3_BUCKET"],
#                }
#            } 
#
#        pool = ["aws:{}:t2.micro".format(os.environ["AWS_REGION"])]        
#
#        self.process (pool) 

if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages
from os import path
import re
import io

module_name = "tcp"
package_name = "tcp-sdk"

deps = [
    'slumber',
    'requests',
    'requests_oauthlib',
    'prettytable'
    ]

def read_file (*relative_path_elements):
    file_path = path.join (path.dirname(__file__), *relative_path_elements)
    return io.open(file_path, encoding='utf8').read().strip()

_version = None
def version ():

    global _version
    if _version:
        return _version
    version_file = read_file(module_name, '_version.py')
    matches = re.search (r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', version_file, re.M)
    if not matches:
        raise RuntimeError ("Unable to find version string in _version.py .")
    _version = matches.group(1)
    return _version

setup (
    name = package_name,

    version = version(),
    description = "Python SDK to query The Cross Product API.",
    long_description_content_type = "text/x-rst",
    long_description = read_file('README.rst'),
    license_files = ("License"),

    install_requires = deps,

    dependency_links = [],

    packages = find_packages(),

    entry_points = {
        'console_scripts': ['tcp_dashboard=tcp.__main__:main']
    },

    test_suite = "tests.get_tests",
)



import sys

def eprint(*args,**kwargs):
    if "debug" in kwargs :
        debug=kwargs["debug"]
        del kwargs['debug']
        if debug : 
            print(*args, file=sys.stderr, **kwargs)
        else:
            print(*args, **kwargs)
    else :
        print(*args, **kwargs)

def create_msg (msg:str, topics=[]):
    from datetime import datetime 
    ts = datetime.utcnow().strftime('%a|%H:%M:%S')
    topics_to_string = ''
    space = ' '
    if ' ' == msg[0]:
        space = ''

    for t in topics:
        func = getattr(t, 'footprint', None)
        if func and callable(func):
            topics_to_string += f'[{t.footprint()}]'
        else:
            topics_to_string += '['+str(t)[:8]+']'

    return f"[{ts}]{topics_to_string}{space}{msg}"

def debug(msg:str, topics=[]):
    eprint (create_msg(msg, topics), debug=True)

def error(msg:str, topics=[]):
    eprint ("[Error]"+create_msg(msg, topics), debug=True)

def warning(msg:str, topics=[]):
    eprint ("[Warning]"+create_msg(msg, topics), debug=True)

